structure Out : sig
  val prop_to_string : NaturalDeduction.prop -> string
  val print_proof : NaturalDeduction.proof -> unit
end = struct
  val string_concat = op^
  val greater_than = op>
  open NaturalDeduction
  val op^ = string_concat
  val op> = greater_than

  fun rule_to_string rule = case rule of
      Hyp u => "hyp_" ^ u
    | TrueI => "TI"
    | FalseE => "FE"
    | AndI => "^I"
    | AndE1 => "^E1"
    | AndE2 => "^E2"
    | OrI1 => "vI1"
    | OrI2 => "vI2"
    | OrE (u,v) => "vE_" ^ u ^ "_" ^ v
    | ImpI u => ">I_" ^ u
    | ImpE => ">E"
  fun atom_to_string atom = case atom of AA => "A" | AB => "B" | AC => "C" | AD => "D" | AE => "E"
  fun prop_to_string prop = case prop of
      PTrue => "T"
    | PFalse => "F"
    | PAtom a => atom_to_string a
    | PAnd (a,b) => (fn (PAtom _|PTrue|PFalse) => prop_to_string a | _ => "(" ^ prop_to_string a ^ ")") a ^ " ^ " ^ (fn (PAtom _|PTrue|PFalse) => prop_to_string b | _ => "(" ^ prop_to_string b ^ ")") b
    | POr (a,b) => "(" ^ prop_to_string a ^ ") v (" ^ prop_to_string b ^ ")"
    | PImp (a as PImp _,b) => "(" ^ prop_to_string a ^ ") > " ^ prop_to_string b
    | PImp (a,b) => prop_to_string a ^ " > " ^ prop_to_string b
  fun judgment_to_string (JTrue p) = prop_to_string p ^ " true"
  fun spaces i = String.implode (List.tabulate (i,(fn _ => #" ")))
  fun pad str width =
    let
      val missing = (width - String.size str)
      val left = missing div 2
      val right = missing - left
    in
      spaces left ^ str ^ spaces right
    end
  fun bar i = String.implode (List.tabulate (i,(fn _ => #"-")))
  fun write_rule rule size = bar (size - String.size rule - 1) ^ rule ^ " "
  fun join_trees (size1,size2) =
    fn([],[]) => []
    | (x::xs,[]) => (x ^ spaces size2) :: join_trees (size1,size2) (xs,[])
    | ([],y::ys) => (spaces size1 ^ y) :: join_trees (size1,size2) ([],ys)
    | (x::xs,y::ys) => (x ^ y) :: join_trees (size1,size2) (xs,ys)
  fun proof_to_tree ((ProofTree (L,r,j))) =
    let
      val L = map proof_to_tree L
      val (size, str, over) = foldl (fn ((size2,str2,over2),(size1,str1,over1)) => (size1+size2,join_trees (size1,size2) (str1,str2),over2)) (0,[],0) L
      val r = " " ^ rule_to_string r ^ " "
      val r_len = String.size r
      val overfill = r_len - over
      val (size,str,over,r) = if overfill > 0 then let val buff = spaces overfill in (size+overfill,map (fn s => s ^ buff) str,over + overfill,r) end else (size,str,over,r ^ spaces (~overfill))
      val j = " " ^ judgment_to_string j ^ " "
      val j_len = String.size j
      val overfill = j_len + over - size
      val (size,str,j) = if overfill > 0 then let val size = size + overfill in (size,map (fn s => (if String.isPrefix "-" s then bar else spaces) overfill ^ s) str,j ^ spaces over) end else (size,str,pad j (size - over) ^ spaces over)
      val j_len = String.size j
      val r = bar (j_len-over) ^ r
    in
      (size,j::r::str,over)
    end
  fun print_proof p = (print "\n"; foldr (fn (s,()) => print (s ^ "\n")) () (#2 (proof_to_tree p)); print "\n")
end